@@ -1,24 +1,27 @@
-/* options.h - global option declarations
- *	Copyright (C) 2002 Miguel Coca.
- *
- * This file is part of GPA
- *
- * GPA is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * GPA is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
- */
+/* options.h - global option declarations.
+   Copyright (C) 2002 Miguel Coca.
+   Copyright (C) 2005, 2008 g10 Code GmbH.
+
+   This file is part of GPA
+  
+   GPA is free software; you can redistribute it and/or modify it
+   under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 3 of the License, or
+   (at your option) any later version.
+  
+   GPA is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
+   License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, see <http://www.gnu.org/licenses/>. */
+
+#include <config.h>
 
 #include <glib.h>
+#include <glib/gstdio.h>
+
 #include "options.h"
 #include "gpa.h"
 #include "gtktools.h"
@@ -28,6 +31,8 @@
 #include <errno.h>
 #include <ctype.h>
 
+#include "confdialog.h" /* gpa_read_configured_keyserver */
+
 /* Internal API */
 static void gpa_options_save_settings (GpaOptions *options);
 static void gpa_options_read_settings (GpaOptions *options);
@@ -41,9 +46,11 @@ static void gpa_options_finalize (GObject *object);
 enum
 {
   CHANGED_UI_MODE,
+  CHANGED_SHOW_ADVANCED_OPTIONS,
   CHANGED_DEFAULT_KEY,
   CHANGED_DEFAULT_KEYSERVER,
   CHANGED_BACKUP_GENERATED,
+  CHANGED_VIEW,
   LAST_SIGNAL
 };
 
@@ -78,6 +85,22 @@ gpa_options_get_type (void)
   return style_type;
 }
 
+
+static void
+make_signal (int signo, GObjectClass *object_class,
+             const char *signame, size_t func_off)
+{
+  g_assert (signo >= 0 && signo < LAST_SIGNAL);
+  signals[signo] = g_signal_new (signame,
+                                 G_TYPE_FROM_CLASS (object_class),
+                                 G_SIGNAL_RUN_FIRST,
+                                 func_off,
+                                 NULL, NULL,
+                                 g_cclosure_marshal_VOID__VOID,
+                                 G_TYPE_NONE, 0);
+}
+
+
 static void
 gpa_options_class_init (GpaOptionsClass *klass)
 {
@@ -88,45 +111,32 @@ gpa_options_class_init (GpaOptionsClass *klass)
   object_class->finalize = gpa_options_finalize;
 
   klass->changed_ui_mode = gpa_options_save_settings;
+  klass->changed_show_advanced_options = gpa_options_save_settings;
   klass->changed_default_key = gpa_options_save_settings;
   klass->changed_default_keyserver = gpa_options_save_settings;
   klass->changed_backup_generated = gpa_options_save_settings;
+  klass->changed_view = gpa_options_save_settings;
 
   /* Signals */
-  signals[CHANGED_UI_MODE] =
-          g_signal_new ("changed_ui_mode",
-                        G_TYPE_FROM_CLASS (object_class),
-                        G_SIGNAL_RUN_FIRST,
-                        G_STRUCT_OFFSET (GpaOptionsClass, changed_ui_mode),
-                        NULL, NULL,
-                        g_cclosure_marshal_VOID__VOID,
-                        G_TYPE_NONE, 0);
-  signals[CHANGED_DEFAULT_KEY] =
-          g_signal_new ("changed_default_key",
-                        G_TYPE_FROM_CLASS (object_class),
-                        G_SIGNAL_RUN_FIRST,
-                        G_STRUCT_OFFSET (GpaOptionsClass, changed_default_key),
-                        NULL, NULL,
-                        g_cclosure_marshal_VOID__VOID,
-                        G_TYPE_NONE, 0);
-  signals[CHANGED_DEFAULT_KEYSERVER] =
-          g_signal_new ("changed_default_keyserver",
-                        G_TYPE_FROM_CLASS (object_class),
-                        G_SIGNAL_RUN_FIRST,
-                        G_STRUCT_OFFSET (GpaOptionsClass,
-                                         changed_default_keyserver),
-                        NULL, NULL,
-                        g_cclosure_marshal_VOID__VOID,
-                        G_TYPE_NONE, 0);
-  signals[CHANGED_BACKUP_GENERATED] =
-          g_signal_new ("changed_backup_generated",
-                        G_TYPE_FROM_CLASS (object_class),
-                        G_SIGNAL_RUN_FIRST,
-                        G_STRUCT_OFFSET (GpaOptionsClass,
-                                         changed_backup_generated),
-                        NULL, NULL,
-                        g_cclosure_marshal_VOID__VOID,
-                        G_TYPE_NONE, 0);
+  make_signal (CHANGED_UI_MODE, object_class,
+               "changed_ui_mode",
+               G_STRUCT_OFFSET (GpaOptionsClass, changed_ui_mode));
+  make_signal (CHANGED_SHOW_ADVANCED_OPTIONS, object_class, 
+               "changed_show_advanced_options",
+               G_STRUCT_OFFSET (GpaOptionsClass,
+                                changed_show_advanced_options));
+  make_signal (CHANGED_DEFAULT_KEY, object_class,
+                "changed_default_key",
+                G_STRUCT_OFFSET (GpaOptionsClass, changed_default_key));
+  make_signal (CHANGED_DEFAULT_KEYSERVER, object_class,
+               "changed_default_keyserver",
+               G_STRUCT_OFFSET (GpaOptionsClass, changed_default_keyserver));
+  make_signal (CHANGED_VIEW, object_class,
+               "changed_view",
+               G_STRUCT_OFFSET (GpaOptionsClass, changed_view));
+  make_signal (CHANGED_BACKUP_GENERATED, object_class,
+               "changed_backup_generated",
+               G_STRUCT_OFFSET (GpaOptionsClass, changed_backup_generated));
 }
 
 static void
@@ -134,9 +144,12 @@ gpa_options_init (GpaOptions *options)
 {
   options->options_file = NULL;
   options->simplified_ui = TRUE;
+  options->show_advanced_options = FALSE;
   options->backup_generated = FALSE;
   options->default_key = NULL;
+  options->default_key_fpr = NULL;
   options->default_keyserver = NULL;
+  options->detailed_view = FALSE;
 }
 
 static void
@@ -145,7 +158,8 @@ gpa_options_finalize (GObject *object)
   GpaOptions *options = GPA_OPTIONS (object);
   
   g_free (options->options_file);
-  g_free (options->default_key);
+  gpgme_key_unref (options->default_key);
+  g_free (options->default_key_fpr);
   g_free (options->default_keyserver);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
@@ -153,26 +167,59 @@ gpa_options_finalize (GObject *object)
 
 /* API */
 
+/* The object */
+static GpaOptions *instance = NULL;
+
 /* Create a new GpaOptions object, reading the options, reading the options
  * from the file given */
-GpaOptions *
-gpa_options_new (const gchar *filename)
+static GpaOptions *
+gpa_options_new (void)
 {
   GpaOptions *options;
   
   options = g_object_new (GPA_OPTIONS_TYPE, NULL);
-  options->options_file = g_strdup (filename);
-  gpa_options_read_settings (options);
 
   return options;
 }
 
+GpaOptions *
+gpa_options_get_instance (void)
+{
+  if (!instance)
+    {
+      instance = gpa_options_new ();
+    }
+  return instance;
+}
+
+
+/* Choose the file the options should be read from and written to.
+ */
+void
+gpa_options_set_file (GpaOptions *options, const gchar *filename)
+{
+  if (options->options_file)
+    {
+      g_free (options->options_file);
+    }
+  options->options_file = g_strdup (filename);
+  gpa_options_read_settings (options);
+}
+
+const gchar *
+gpa_options_get_file (GpaOptions *options)
+{
+  return options->options_file;
+}
+
 /* Set whether the ui should be in simplified mode */
 void
 gpa_options_set_simplified_ui (GpaOptions *options, gboolean value)
 {
+  int change = (!options->simplified_ui != !value);
   options->simplified_ui = value;
-  g_signal_emit (options, signals[CHANGED_UI_MODE], 0);
+  if (change)
+    g_signal_emit (options, signals[CHANGED_UI_MODE], 0);
 }
 
 gboolean
@@ -181,69 +228,103 @@ gpa_options_get_simplified_ui (GpaOptions *options)
   return options->simplified_ui;
 }
 
+/* Set whether the preference dialog shows the advanced options. */
+void
+gpa_options_set_show_advanced_options (GpaOptions *options, gboolean value)
+{
+  int change = (!options->show_advanced_options != !value);
+  options->show_advanced_options = value;
+  if (change)
+    g_signal_emit (options, signals[CHANGED_SHOW_ADVANCED_OPTIONS], 0);
+}
+
+gboolean
+gpa_options_get_show_advanced_options (GpaOptions *options)
+{
+  return options->show_advanced_options;
+}
+
 /* Choose the default key */
 void
-gpa_options_set_default_key (GpaOptions *options, const gchar *fpr)
+gpa_options_set_default_key (GpaOptions *options, gpgme_key_t key)
 {
   if (options->default_key)
     {
-      g_free (options->default_key);
+      gpgme_key_unref (options->default_key);
+      options->default_key = NULL;
+    }
+  if (key)
+    {
+      gpgme_key_ref (key);
+      options->default_key = key;
+      options->default_key_fpr = g_strdup (key->subkeys->fpr);
     }
-  options->default_key = g_strdup (fpr);
   g_signal_emit (options, signals[CHANGED_DEFAULT_KEY], 0);
 }
 
-const gchar *
+
+gpgme_key_t
 gpa_options_get_default_key (GpaOptions *options)
 {
   return options->default_key;
 }
 
+
 /* Return the default key gpg would use, or at least a first
  * approximation. Currently this means the first secret key in the keyring.
  * If there's no secret key at all, return NULL
  */
-static gchar *
+static gpgme_key_t
 determine_default_key (void)
 {
-  GpgmeKey key;
-  GpgmeError err;
-  gchar * fpr = NULL;
-  err = gpgme_op_keylist_start (ctx, NULL, 1);
-  if( err != GPGME_No_Error )
-    gpa_gpgme_error (err);
-  err = gpgme_op_keylist_next (ctx, &key);
-  if( err == GPGME_No_Error )
-    {
-      /* Copy the ID and release the key, since it's no longer needed */
-      fpr = g_strdup (gpgme_key_get_string_attr (key, GPGME_ATTR_FPR, 
-                                                 NULL, 0 ));
-      gpgme_key_release (key);
-      err = gpgme_op_keylist_end (ctx);
-      if( err != GPGME_No_Error )
-        gpa_gpgme_error (err); 
-    }
-  else if( err == GPGME_EOF )
-    /* No secret keys found */
-    fpr = NULL;
-  else
-    gpa_gpgme_error (err);
+  gpgme_key_t key = NULL;
+  gpg_error_t err;
+  gpgme_ctx_t ctx;
 
-  return fpr;
+  ctx = gpa_gpgme_new ();
+  if (!ctx)
+    return NULL;
+
+  err = gpgme_op_keylist_start (ctx, NULL, 1);
+  if (err)
+    {
+      gpa_gpgme_error (err); 	/* Or should we use
+				   gpa_gpgme_warning()?  -mo*/
+    }
+  else
+    {
+      err = gpgme_op_keylist_next (ctx, &key);
+      if (!err)
+	{
+	  /* Got a new key. */
+	  err = gpgme_op_keylist_end (ctx);
+	  if (gpg_err_code (err) != GPG_ERR_NO_ERROR)
+	    {
+	      gpa_gpgme_warning (err); 
+	    }
+	}
+      else if (gpg_err_code (err) != GPG_ERR_EOF)
+	{
+	  gpa_gpgme_error (err); 	/* Or should we use
+					   gpa_gpgme_warning()?  -mo*/
+	}
+    }
+
+  gpgme_release (ctx);
+  return key;
 }
 
 void
 gpa_options_update_default_key (GpaOptions *options)
 {
   gboolean update = FALSE;
-  GpgmeKey key = NULL;
+  gpgme_key_t key = NULL;
+  gpgme_ctx_t ctx = gpa_gpgme_new ();
 
-  if (!options->default_key)
-    {
-      update = TRUE;
-    }
-  else if (gpgme_get_key (ctx, options->default_key, &key, TRUE, TRUE) == 
-           GPGME_EOF)
+  if (! options->default_key_fpr)
+    update = TRUE;
+  else if (gpg_err_code (gpgme_get_key (ctx, options->default_key_fpr, 
+					&key, TRUE)) == GPG_ERR_EOF)
     {
       gpa_window_error (_("The private key you selected as default is no "
                           "longer available.\n"
@@ -251,18 +332,25 @@ gpa_options_update_default_key (GpaOptions *options)
                           "key automatically."), NULL);
       update = TRUE;
     }
+  else
+    {
+      /* Use the listed key if there is no previous one (at initialization).
+       */
+      if (!options->default_key)
+	options->default_key = key;
+      else
+	gpgme_key_unref (key);
+    }
 
   if (update)
     {
-      gchar *fpr = determine_default_key ();
-      gpa_options_set_default_key (options, fpr);
-      g_free (fpr);
+      key = determine_default_key ();
+      gpa_options_set_default_key (options, key);
+      if (key)
+	gpgme_key_unref (key);
     }
 
-  if (key)
-    {
-      gpgme_key_unref (key);
-    }
+  gpgme_release (ctx);
 }
 
 /* Specify the default keyserver */
@@ -283,6 +371,21 @@ gpa_options_get_default_keyserver (GpaOptions *options)
   return options->default_keyserver;
 }
 
+/* Remember whether using brief or detailed view */
+void
+gpa_options_set_detailed_view (GpaOptions *options, gboolean value)
+{
+  options->detailed_view = value;
+  g_signal_emit (options, signals[CHANGED_VIEW], 0);
+}
+
+gboolean
+gpa_options_get_detailed_view (GpaOptions *options)
+{
+  return options->detailed_view;
+}
+
+
 /* Remember whether the default key has already been backed up */
 void
 gpa_options_set_backup_generated (GpaOptions *options, gboolean value)
@@ -297,20 +400,13 @@ gpa_options_get_backup_generated (GpaOptions *options)
   return options->backup_generated;
 }
 
-/* Destroy the GpaOptions object */
-void
-gpa_options_destroy (GpaOptions *options)
-{
-  g_object_run_dispose (G_OBJECT (options));
-}
-
 static void
 gpa_options_save_settings (GpaOptions *options)
 {
   FILE *options_file;
   
   g_assert(options->options_file != NULL);
-  options_file =  fopen (options->options_file, "w");
+  options_file = g_fopen (options->options_file, "w");
   if (!options_file)
     {
       g_warning("%s: %s", options->options_file, strerror (errno));
@@ -319,12 +415,14 @@ gpa_options_save_settings (GpaOptions *options)
     {
       if (options->default_key)
         {
-          fprintf (options_file, "default-key %s\n", options->default_key);
+          fprintf (options_file, "default-key %s\n",
+		   options->default_key->subkeys->fpr);
         }
       if (options->default_keyserver)
         {
-          fprintf (options_file, "keyserver %s\n", 
-                   options->default_keyserver);
+          /* we do not write the keyserver anymore to gpg.conf.  */
+/*           fprintf (options_file, "keyserver %s\n",  */
+/*                    options->default_keyserver); */
         }
       if (options->backup_generated)
         {
@@ -334,15 +432,28 @@ gpa_options_save_settings (GpaOptions *options)
         {
           fprintf (options_file, "%s\n", "advanced-ui");
         }
+      if (options->show_advanced_options)
+        {
+          fprintf (options_file, "%s\n", "show-advanced-options");
+        }
+      if (options->detailed_view)
+        {
+          fprintf (options_file, "%s\n", "detailed-view");
+        }
       fclose (options_file);
     }
+
+  /* Write the keyserver to the backend.  */
+  if (options->default_keyserver)
+    gpa_store_configured_keyserver (options->default_keyserver);
 }
 
+
 static gboolean
 read_next_word (FILE *file, char *buffer, int size)
 {
   int i = 0;
-  char c;
+  int c;
 
   buffer[0] = '\0';
   /* Skip leading whitespace */
@@ -381,7 +492,7 @@ gpa_options_read_settings (GpaOptions *options)
   FILE *options_file;
   
   g_assert(options->options_file != NULL);
-  options_file =  fopen (options->options_file, "r");
+  options_file = g_fopen (options->options_file, "r");
   if (!options_file)
     {
       /* If the config file just doesn't exist, it's not an error */
@@ -419,13 +530,22 @@ gpa_options_read_settings (GpaOptions *options)
                 {
                   options->simplified_ui = FALSE;
                 }
+              else if (g_str_equal (next_word, "show-advanced-options"))
+                {
+                  options->show_advanced_options = TRUE;
+                }
+              else if (g_str_equal (next_word, "detailed-view"))
+                {
+                  options->detailed_view = TRUE;
+                }
               break;
             case PARSE_OPTIONS_STATE_HAVE_KEY:
-              options->default_key = g_strdup (next_word);
+              options->default_key_fpr = g_strdup (next_word);
               state = PARSE_OPTIONS_STATE_START;
               break;
             case PARSE_OPTIONS_STATE_HAVE_KEYSERVER:
-              options->default_keyserver = g_strdup (next_word);
+              /* We do not use the keyserver item from the gpa.conf anymore. */
+              /* options->default_keyserver = g_strdup (next_word); */
               state = PARSE_OPTIONS_STATE_START;
               break;
             default:
@@ -435,4 +555,8 @@ gpa_options_read_settings (GpaOptions *options)
         }
       fclose (options_file);
     }
+
+  /* Read the keyserver from the abckend.  */
+  g_free (options->default_keyserver);
+  options->default_keyserver = gpa_load_configured_keyserver ();
 }
